'''Command line interface'''
# pylint: disable=too-many-arguments,invalid-name,too-many-locals

import os
import re
import sys

import click
from click import Context

from textwatermark import __version__
from textwatermark.config import settings
from textwatermark.defines import WMMethod, WMMode
from textwatermark.main import TextWatermark
from textwatermark.template_type import WMTemplateType

# from textwatermark.log import init_log


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('-V', '--version', is_flag=True, help='Show version and exit.')
@click.option('-v', '--verbose', is_flag=True, help='Show more info.')
# If it's true, it will override `settings.VERBOSE`
@click.option('--debug', is_flag=True, help='Enable debug.')
# If it's true, it will override `settings.DEBUG`
def main(ctx: Context, version: str, verbose: bool, debug: bool):
    '''Main commands'''
    if version:
        click.echo(__version__)
    elif ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
    else:
        if verbose:
            settings.set('VERBOSE', True)
        if debug:
            settings.set('DEBUG', True)


@main.command()
@click.option('-f', '--text-file',  type=str, required=False,
              help='Text file waiting for watermarking')
@click.option('-o', '--out-file',  type=str, required=False,
              help='Watermarked file to be saved')
@click.option('-m', '--wm-mode',  type=str, required=True,
              help='Watermark mode value in defines.WMMode')
@click.option('-t', '--template-type',  type=str, required=True,
              help='Template type in templates')
@click.option('-x', '--wm-max',  type=str, required=True,
              help='Max value or string of the watermark')
@click.option('-w', '--wm-str',  type=str, required=True,
              help='Watermark string')
@click.option('-b', '--wm-base',  type=int, default=0, required=False,
              help='Base conversion of watermark string')
@click.option('-k', '--template-chars-key', type=str, default='', required=False,
              help='Key of template confusables chars')
@click.option('-l', '--wm-loop',  is_flag=True,
              help='If True then inserts watermark in a loop, Defaults to False')
@click.option('-i', '--start-at',  type=int, default=0, required=False,
              help='Index of where the watermark will be inserted. Defaults to 0.')
@click.option('-e', '--export-params', is_flag=True,
              help='If True then export watermark params')
@click.option('-n', '--no-flag-bit', is_flag=True,
              help='If True then do not add a flag bit to watermark')
def insert(text_file: str, out_file: str, wm_mode: str, template_type: str, wm_max: int,
           wm_str: str, wm_base: int, template_chars_key: str, wm_loop: bool, start_at: int,
           export_params: bool, no_flag_bit: bool):
    '''Insert watermark to text

    Examples:

    Insert watermark to text file:

    `textwatermark -v insert -f './tests/text/1.txt' -m ALPHA_NUMERICAL
    -t HOMOGRAPH_NUMBERS -x 999999999 -w 123456789 `

    Export params to out_file:

    `textwatermark -v insert -m ALPHA_NUMERICAL -t HOMOGRAPH_NUMBERS
    -x 999999999 -w 123456789 -e -o 'out.txt'`
    '''
    verbose = settings.get('VERBOSE')

    template_type = WMTemplateType[template_type]
    if template_type.value.method == WMMethod.DECORATE_EACH_CHAR and \
            template_chars_key == '':
        raise ValueError('template_chars_key is required')

    if re.fullmatch('^\\\\u[0-9]{4}$', template_chars_key) is not None:
        template_chars_key = chr(int(template_chars_key[2:], 16))

    # # init
    wm = TextWatermark(WMMode[wm_mode].value, wm_base,
                       start_at, wm_loop, wm_flag_bit=not no_flag_bit)
    wm.set_tpl_type(template_type, template_chars_key)
    wm.set_wm_max(wm_max=wm_max)

    if export_params:
        # wm.tpl_type = ''
        if out_file:
            out_file_path = os.path.abspath(out_file)
            with open(out_file_path, 'w', encoding="utf-8") as f:
                f.write(wm.export_params())
            if verbose:
                print('Export params save to output file: '+out_file_path)
            else:
                print('ok')
        else:
            print(wm.export_params())
        return

    text_file = os.path.abspath(text_file)
    wm.set_text_file(text_file)
    wm_text = wm.insert_watermark(wm_str)
    if out_file:
        out_file_path = os.path.abspath(out_file)
        wm.save_to_file(wm_text, out_file_path)
        if verbose:
            print('Save watermarked text to output file: '+out_file_path)
            print('Orgin text length is: '+str(len(wm.text)))
            print('Watermarked text length is: '+str(len(wm_text)))
        else:
            print('ok')
    else:
        print(wm_text)


@main.command()
@click.option('-f', '--wm-text-file',  type=str, required=False,
              help='Text file already be watermarked')
@click.option('-b', '--wm-binary',  type=str, required=False,
              help='Watermark string in binary')
@click.option('-p', '--params-json',  type=str, required=True,
              help='Param json when watermarking text')
@click.option('-F', '--force-check-version',  is_flag=True,
              help='Force to check versions between params and library')
def retrieve(wm_text_file: str, wm_binary: str, params_json: str,
             force_check_version: bool):
    '''Retrieve watermark from watermarked text

    Examples:

    `textwatermark retrieve -f ./out.txt -p '{the param json string export by  
    command:insert and option:--export-params}'`
    '''
    verbose = settings.get('VERBOSE')

    if not wm_text_file and not wm_binary:
        raise ValueError(
            'No watermarked text file or watermark binary string set')

    if wm_binary:
        wm_out_str = TextWatermark.retrieve_watermark_from_bin(
            wm_binary, params_json, force_check_version)

    else:
        wm_text_file = os.path.abspath(wm_text_file)
        if not os.path.exists(wm_text_file):
            raise ValueError(f'ERROR: file {wm_text_file} does not exist')
        try:
            with open(wm_text_file, 'r', encoding='utf-8') as f:
                wm_text = f.read()
        except OSError as err:
            print(
                f'ERROR: cannot read file {wm_text_file}, err is {err.strerror}')
            sys.exit()
        wm_out_str = TextWatermark.retrieve_watermark(wm_text, params_json,
                                                      force_check_version)

    if verbose:
        print(f'The retrieved watermark is: {wm_out_str}')
    else:
        print(wm_out_str)



'''
Insert the selected character after the text to represent a 1 in binary, otherwise 0.

The characters are from [combining_chars][], [invisible_chars][] and [whitespace_chars][].

Warning: confusables_chars_key is needed
    You can choose one of the characters to insert the watermark

Info: Differents between combining_chars, invisible_chars and whitespace_chars
    [combining_chars][]: Small but recognizable difference in characters 
    before and after watermarking

    [invisible_chars][]: You can't see the difference between before 
    and after inserting the watermark

    [whitespace_chars][]: There is a small but recognizable difference in 
    the spacing of characters before and after watermarking

'''

from textwatermark.defines import WMMethod
from textwatermark.templates import (combining_chars, invisible_chars,
                                     whitespace_chars)

CONFUSABLES_CHARS_LIST = combining_chars.CONFUSABLES_CHARS + \
    invisible_chars.CONFUSABLES_CHARS + whitespace_chars.CONFUSABLES_CHARS + \
    ['\u007F']

CONFUSABLES_CHARS = dict(
    zip(CONFUSABLES_CHARS_LIST, CONFUSABLES_CHARS_LIST))

method = WMMethod.APPEND_AS_BINARY



'''
The template will generate a watermark into many types of space chars 
and then insert it into the specified location.

The best usage scenario for this template is to insert the watermark on 
an empty line of text or after some lines in the code.
'''

from textwatermark.defines import WhiteSpaceChars, WMMethod

# 空格字符
CONFUSABLES_CHARS = [
    # WhiteSpaceChars.SPACE_1_4.value,
    WhiteSpaceChars.NO_BREAK_SPACE_1_4.value,
    WhiteSpaceChars.EN_QUAD_1_2.value,
    WhiteSpaceChars.EM_QUAD_1.value,
    WhiteSpaceChars.EN_SPACE_1_2.value,
    WhiteSpaceChars.EM_SPACE_1.value,
    WhiteSpaceChars.THREE_PER_EM_SPACE_1_3.value,
    WhiteSpaceChars.FOUR_PER_EM_SPACE_1_4.value,
    WhiteSpaceChars.SIX_PER_EM_SPACE_1_6.value,
    WhiteSpaceChars.FIGURE_SPACE_LIKE_DIGITS.value,
    WhiteSpaceChars.PUNCTUATION_SPACE_LIKE_DOT.value,
    WhiteSpaceChars.THIN_SPACE_1_6.value,
    WhiteSpaceChars.HAIR_SPACE_1_8.value,
    WhiteSpaceChars.NARROW_NO_BREAK_SPACE_1_6.value,
    WhiteSpaceChars.MEDIUM_MATHEMATICAL_SPACE_2_9.value,
]


method = WMMethod.INSERT_INTO_POSITION



'''
HTML empty tags template can only be used when the text can be displayed in HTML mode.
Such as WEB page, Blog, EMail, etc. 

All the tags below are inline tags.

Note: Some of HTML editor will remove empty or outside of list tags.
    Therefore, this template is not a relatively stable template, 
    and it is not recommended to use it in most cases.
'''

from textwatermark.defines import WMMethod

# 空格字符
CONFUSABLES_CHARS = [
    '<a></a>',
    '<abbr></abbr>',
    '<b></b>',
    '<bdo></bdo>',
    '<big></big>',
    '<cite></cite>',
    '<code></code>',
    '<dfn></dfn>',
    '<em></em>',
    '<i></i>',
    '<kbd></kbd>',
    '<var></var>',
    '<samp></samp>',
    '<span></span>',
    '<strong></strong>',
    '<small></small>',
    '<sub></sub>',
    '<sup></sup>',
    '<tt></tt>',
    '<u></u>',
]

method = WMMethod.INSERT_INTO_POSITION



'''Class of Conversion'''
# pylint: disable=too-many-instance-attributes

import re

import numpy as np
from bitstring import BitArray

from textwatermark.defines import WMMode


class WMConversion:
    '''
    A class to insert watermark into plain text

    Example:
    --------
    ```py
    wm_base = 0
    wm_str = '123'
    wmc = WMConversion(WMMode.REAL_NUMBER, wm_base)
    wm_final = wmc.wm_convert_to_arbitrary_base(wm_str)
    wm_out_str = wmc.wm_restore_from_arbitrary_base(wm_final)
    assert wm_str == wm_out_str
    print(wmc.calc_max_wm_length(wm_str+wm_str), wm_final, wm_str, wm_out_str)
    ```
    '''

    def __init__(self, wm_mode: WMMode, wm_base: int):
        '''Class TextWatermark init

        If debug is set to True, the watermark processing will be printed.

        Args:
            wm_mode (WMMode): Set watermark encoding mode. 
            wm_base (int): Watermark's conversion base

        Raises:
            ValueError: If `base` is not in the range [2, 36]
        '''

        self.wm_mode = wm_mode
        '''Setting different watermark modes will effectively reduce the size of 
        the watermark string.'''

        if wm_base < 2 or wm_base > 36:
            raise ValueError(
                f'watermark convert base must be between 2 and 36: {wm_base}')

        self.wm_base = wm_base
        '''watermark conversion base must be between 2 and 36'''

    def calc_max_wm_length(self, wm_max: str):
        '''Calculate the watermark length in target base

        This method usually be used with `wm.set_fixed_wm_length` method.
        This method will calculate the length of the longest (or largest) 
        watermark string in the input string. Then call `wm.set_fixed_wm_length` 
        method set the return value to wm.wm_fixed_len.

        **It should be noted that:** *the length calculated by this method will 
        add 1 to the original length. The purpose is to quickly locate the 
        starting position of the watermark string when filling the watermark 
        string with 0, so as to ensure that the correct value is obtained and 
        restored. *

        Args:
            wm_max (str): Maximum watermark string in target base

        Returns:
            (int): Returns the length of the watermark string in the specified base
        '''

        wm_final = self.wm_convert_to_arbitrary_base(wm_max)
        return len(wm_final)

    def wm_convert_to_arbitrary_base(self, wm_str: str):
        '''Convert watermark to arbitrary base string

        Args:
            wm_str (str): Watermark string

        Returns:
            (str): Returns watermark string in arbitrary base

        Raises:
            ValueError: if `wm_str` type is not match `wm_mode`
            ValueError: if invalid watermark `wm_mode`
            ValueError: if invalid watermark `wm_base`

        '''

        if not isinstance(wm_str, str):
            wm_str = str(wm_str)

        wm_bin = ''
        match self.wm_mode:
            case WMMode.REAL_NUMBER:
                if not re.fullmatch(r'[0-9]+', wm_str):
                    raise ValueError(
                        f'ERROR: invalid watermark wm_mode: {wm_str} is not numerical ')
                wm_bin = str(bin(int(wm_str)))[2:]
            case WMMode.LETTERS_LOWER_CASE:
                if not re.fullmatch(r'[a-z]+', wm_str):
                    raise ValueError(
                        f'ERROR: invalid watermark wm_mode: {wm_str} are not lower case letters ')
                wm_bin = ''.join(
                    format(ord(i)-ord('a')+1, '05b') for i in wm_str)
            case WMMode.LETTERS_UPPER_CASE:
                if not re.fullmatch(r'[A-Z]+', wm_str):
                    raise ValueError(
                        f'ERROR: invalid watermark wm_mode: {wm_str} are not upper case letters ')
                wm_bin = ''.join(
                    format(ord(i)-ord('A')+1, '05b') for i in wm_str)
            case WMMode.LETTERS_MIXED_CASE:
                if not re.fullmatch(r'[a-zA-Z]+', wm_str):
                    raise ValueError(
                        f'ERROR: invalid watermark wm_mode: {wm_str} are not mixed case letters ')
                wm_bin = ''.join(
                    format(ord(i)-ord('A')+1, '06b') for i in wm_str)
            case WMMode.ALPHA_NUMERICAL:
                if not re.fullmatch(r'[0-9a-zA-Z]+', wm_str):
                    raise ValueError(
                        f'ERROR: invalid watermark wm_mode: {wm_str} are not mixed case letters ')
                for i in wm_str:
                    ord_i = ord(i)
                    if ord('0') <= ord_i <= ord('9'):
                        wm_bin += format(ord_i - ord('0') + 1, '06b')
                    elif ord('A') <= ord_i <= ord('Z'):
                        wm_bin += format(ord_i - ord('A') + 1 + 10, '06b')
                    elif ord('a') <= ord_i <= ord('z'):
                        wm_bin += format(ord_i - ord('a') + 1 + 36, '06b')
            case WMMode.ALPHA_NUMERICAL_SPECIAL:
                if not re.fullmatch(r'[!-~]+', wm_str):
                    raise ValueError(
                        f'ERROR: invalid watermark wm_mode: {wm_str} are not mixed case letters ')
                wm_bin = ''.join(format(ord(i), '07b') for i in wm_str)
            case WMMode.UNICODE:
                wm_bin = BitArray(wm_str.encode('UTF-8')).bin
            case _:
                raise ValueError(
                    f'ERROR: invalid watermark wm_mode: {self.wm_mode}')

        # convert wm_bin to arbitrary base string
        if self.wm_base < 2 or self.wm_base > 36:
            raise ValueError(
                f'Error: Unsupported base value. Must be between 2 and 16: {self.wm_base}')
        num = np.base_repr(int(wm_bin, 2), self.wm_base)
        return str(num)

    def _get_str_from_bin(self, wm_bin: str, bin_len: int, offset: int):
        '''Generate the watermark string from binary string.

        Args:
            wm_bin (str): watermark binary string
            bin_len (int): The number of bytes representing a watermark character
            offset (int): The number of leading bits of the binary string

        Returns:
            (str): restored watermark string
        '''

        wm_str = ''
        if len(wm_bin) % bin_len != 0:
            wm_bin = wm_bin.zfill(bin_len - len(wm_bin) %
                                  bin_len + len(wm_bin))

        for i in range(0, len(wm_bin), bin_len):
            wm_str += chr(int(wm_bin[i:i + bin_len], 2) + offset)
        return wm_str

    def wm_restore_from_arbitrary_base(self, wm_final: str):
        '''Restore watermark from arbitrary base string

        Args:
            wm_final (str): Arbitrary base watermark string

        Returns:
            (str): restored watermark string

        Raises:
            Error: If Invalid WMMode

        '''

        # convert wm_bin from arbitrary base string
        wm_bin = str(bin(int(wm_final, self.wm_base)))[2:]

        wm_str = ''
        match self.wm_mode:
            case WMMode.REAL_NUMBER:
                wm_str = str(int(wm_bin, 2))
            case WMMode.LETTERS_LOWER_CASE:
                wm_str = self._get_str_from_bin(wm_bin, 5, ord('a')-1)
            case WMMode.LETTERS_UPPER_CASE:
                wm_str = self._get_str_from_bin(wm_bin, 5, ord('A')-1)
            case WMMode.LETTERS_MIXED_CASE:
                wm_str = self._get_str_from_bin(wm_bin, 6, ord('A')-1)
            case WMMode.ALPHA_NUMERICAL:
                if len(wm_bin) % 6 != 0:
                    wm_bin = wm_bin.zfill(6 - len(wm_bin) % 6 + len(wm_bin))
                for i in range(0, len(wm_bin), 6):
                    num = int(wm_bin[i:i + 6], 2)
                    if num <= 10:
                        wm_str += chr(num + ord('0') - 1)
                    elif 10 < num <= 36:
                        wm_str += chr(num + ord('A') - 1 - 10)
                    elif 36 < num <= 62:
                        wm_str += chr(num + ord('a') - 1 - 36)
            case WMMode.ALPHA_NUMERICAL_SPECIAL:
                wm_str = self._get_str_from_bin(wm_bin, 7, 0)
            case WMMode.UNICODE:
                wm_str = BitArray(bin=wm_bin).bytes.decode(encoding='UTF-8')
            case _:
                raise ValueError(f'ERROR: Invalid WMMode: {self.wm_mode}')

        return wm_str



'''
Font style template can only be used when the text can be displayed in HTML mode.
Such as WEB page, Blog, EMail, etc. 
You can customize your own style pairs following the template format.

Note: Need to wrap surrounding elements
    It should be noted that all text needs to be wrapped with a layer of elements in 
    order to obtain a unified style display

Warning: confusables_chars_key is needed
    You can choose one of the key to insert the watermark

'''

from textwatermark.common import add_head_n_tial_to_dict
from textwatermark.defines import WMMethod

# 组合字符，如音标等
CONFUSABLES_CHARS_START_WITH = '<span style="'
CONFUSABLES_CHARS_END_WITH = ';">{char}</span>'
CONFUSABLES_CHARS = {
    'text-decoration-underline': ['text-decoration:none', 'text-decoration:underline'],
    'text-decoration-line-through': ['text-decoration: none', 'text-decoration: line-through'],
    'text-decoration-overline': ['text-decoration: none', 'text-decoration: overline'],
    'background-color-white-gray-1': ['background-color:#fff', 'background-color:#eee'],
    'background-color-white-gray-2': ['background-color:#fff', 'background-color:#ddd'],
    'background-color-white-red-1': ['background-color:#fff', 'background-color:#eff'],
    'background-color-white-red-2': ['background-color:#fff', 'background-color:#dff'],
    'background-color-white-green-1': ['background-color:#fff', 'background-color:#fef'],
    'background-color-white-green-2': ['background-color:#fff', 'background-color:#fdf'],
    'background-color-white-blue-1': ['background-color:#fff', 'background-color:#ffe'],
    'background-color-white-blue-2': ['background-color:#fff', 'background-color:#ffd'],
    'background-color-black-gray-1': ['background-color:#000', 'background-color:#111'],
    'background-color-black-gray-2': ['background-color:#000', 'background-color:#222'],
    'background-color-black-red-1': ['background-color:#000', 'background-color:#100'],
    'background-color-black-red-2': ['background-color:#000', 'background-color:#200'],
    'background-color-black-green-1': ['background-color:#000', 'background-color:#010'],
    'background-color-black-green-2': ['background-color:#000', 'background-color:#020'],
    'background-color-black-blue-1': ['background-color:#000', 'background-color:#001'],
    'background-color-black-blue-2': ['background-color:#000', 'background-color:#002'],
    'border-top-white-dashed-1': ['border-top: ;', 'border-top: 1px dashed #eee;'],
    'border-top-white-dashed-2': ['border-top: ;', 'border-top: 1px dashed #ddd;'],
    'border-top-white-dashed-3': ['border-top: ;', 'border-top: 1px dashed #ccc;'],
    'border-bottom-white-dashed-1': ['border-bottom: ;', 'border-bottom: 1px dashed #eee;'],
    'border-bottom-white-dashed-2': ['border-bottom: ;', 'border-bottom: 1px dashed #ddd;'],
    'border-bottom-white-dashed-3': ['border-bottom: ;', 'border-bottom: 1px dashed #ccc;'],
    'border-left-white-dashed-1': ['border-left: ;', 'border-left: 1px dashed #eee;'],
    'border-left-white-dashed-2': ['border-left: ;', 'border-left: 1px dashed #ddd;'],
    'border-left-white-dashed-3': ['border-left: ;', 'border-left: 1px dashed #ccc;'],
    'border-right-white-dashed-1': ['border-right: ;', 'border-right: 1px dashed #eee;'],
    'border-right-white-dashed-2': ['border-right: ;', 'border-right: 1px dashed #ddd;'],
    'border-right-white-dashed-3': ['border-right: ;', 'border-right: 1px dashed #ccc;'],
    'border-top-white-solid-1': ['border-top: ;', 'border-top: 1px solid #eee;'],
    'border-top-white-solid-2': ['border-top: ;', 'border-top: 1px solid #ddd;'],
    'border-top-white-solid-3': ['border-top: ;', 'border-top: 1px solid #ccc;'],
    'border-bottom-white-solid-1': ['border-bottom: ;', 'border-bottom: 1px solid #eee;'],
    'border-bottom-white-solid-2': ['border-bottom: ;', 'border-bottom: 1px solid #ddd;'],
    'border-bottom-white-solid-3': ['border-bottom: ;', 'border-bottom: 1px solid #ccc;'],
    'border-left-white-solid-1': ['border-left: ;', 'border-left: 1px solid #eee;'],
    'border-left-white-solid-2': ['border-left: ;', 'border-left: 1px solid #ddd;'],
    'border-left-white-solid-3': ['border-left: ;', 'border-left: 1px solid #ccc;'],
    'border-right-white-solid-1': ['border-right: ;', 'border-right: 1px solid #eee;'],
    'border-right-white-solid-2': ['border-right: ;', 'border-right: 1px solid #ddd;'],
    'border-right-white-solid-3': ['border-right: ;', 'border-right: 1px solid #ccc;'],
    'font-style-italic': ['font-style:normal', 'font-style:italic']

}

add_head_n_tial_to_dict(
    CONFUSABLES_CHARS, CONFUSABLES_CHARS_START_WITH, CONFUSABLES_CHARS_END_WITH)

method = WMMethod.DECORATE_EACH_CHAR



'''
This template will replace the regular space char with the other width of space chars.

Since spaces are rare in Chinese, this template is mainly suitable for English text.
'''

from textwatermark.defines import WhiteSpaceChars, WMMethod

# 空格字符
CONFUSABLES_CHARS = {
    '\u0020': [
        WhiteSpaceChars.SPACE_1_4.value,
        # WhiteSpaceChars.NO_BREAK_SPACE_1_4.value,
        # WhiteSpaceChars.EN_QUAD_1_2.value,
        # WhiteSpaceChars.EM_QUAD_1.value,
        WhiteSpaceChars.EN_SPACE_1_2.value,
        WhiteSpaceChars.EM_SPACE_1.value,
        WhiteSpaceChars.THREE_PER_EM_SPACE_1_3.value,
        # WhiteSpaceChars.FOUR_PER_EM_SPACE_1_4.value,
        WhiteSpaceChars.SIX_PER_EM_SPACE_1_6.value,
        # WhiteSpaceChars.FIGURE_SPACE_LIKE_DIGITS.value,
        # WhiteSpaceChars.PUNCTUATION_SPACE_LIKE_DOT.value,
        # WhiteSpaceChars.THIN_SPACE_1_6.value,
        WhiteSpaceChars.HAIR_SPACE_1_8.value,
        # WhiteSpaceChars.NARROW_NO_BREAK_SPACE_1_6.value,
        # WhiteSpaceChars.MEDIUM_MATHEMATICAL_SPACE_2_9.value,
    ]
}

method = WMMethod.FIND_AND_REPLACE



'''Some common functions here'''


def add_head_n_tial_to_dict(the_dict: dict, head: str, tail: str):
    '''Add head and tail to a dict, will directly effect to origin dict.

    Args:
        the_dict (dict): dict to add head and tail to
        head (str): head string to add
        tail (str): tail string to add

    '''
    for _, ival in the_dict.items():
        for jkey, _ in enumerate(ival):
            ival[jkey] = head + ival[jkey] + tail
